#define CHECKSUM_START 0x1000
#define CHECKSUM_LENGTH 0x100000L
#define CHECKSUM_HEADERPOS 0x10
#define CHECKSUM_END (CHECKSUM_START + CHECKSUM_LENGTH)
#define CHECKSUM_STARTVALUE 0xf8ca4ddc//makesram routine
#define ROL(i, b) (((i)<<(b)) | ((i)>>(32-(b))))
#define max2(a, b) ( (a)>(b) ? (a) : (b) )
#define min2(a, b) ( (a)<(b) ? (a) : (b) )
#define BYTES2LONG(b, s) ( (((b)[0^(s)] & 0xffL) << 24) | \
                           (((b)[1^(s)] & 0xffL) << 16) | \
                           (((b)[2^(s)] & 0xffL) <<  8) | \
                           (((b)[3^(s)] & 0xffL)) )

#define LONG2BYTES(l, b, s)  (b)[0^(s)] = ((l)>>24)&0xff; \
                             (b)[1^(s)] = ((l)>>16)&0xff; \
                             (b)[2^(s)] = ((l)>> 8)&0xff; \
                             (b)[3^(s)] = ((l)    )&0xff;

#define HEADER_MAGIC 0x80371240

int n64chksum(char *name,long hsize,int swapped)
{
//   ROM Checksum Routine Courtesy of:
//   chksum64 V1.2, a program to calculate the ROM checksum of Nintendo64 ROMs.
//   Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)

char readchunk3[32768];
    unsigned long i;
    unsigned long c1, k1, k2;
    unsigned long t1, t2, t3, t4;
    unsigned long t5, t6;
    unsigned int n;
    long clen = CHECKSUM_LENGTH;
    long rlen;
	FILE *fh;
	unsigned long sum1, sum2;
    rlen = fsize(name,hsize) - CHECKSUM_START;
    t1 = CHECKSUM_STARTVALUE;
    t2 = CHECKSUM_STARTVALUE;
    t3 = CHECKSUM_STARTVALUE;
    t4 = CHECKSUM_STARTVALUE;
    t5 = CHECKSUM_STARTVALUE;
    t6 = CHECKSUM_STARTVALUE;

	if(fsize(name,hsize)<0x0100000)//0x0101000
	{
		return(-1);//rom is too short
	}
fhexdump(name,CHECKSUM_HEADERPOS+hsize,8);
	if(!(fh=fopen(name,"r+b")))return(-1);
	fseek(fh, CHECKSUM_START+hsize, SEEK_SET);

for( ;; ) 
{
	if( rlen > 0 ) 
	{
        	n = fread(readchunk3, 1, min2(sizeof(readchunk3), clen), fh);
        	if( (n & 0x03) != 0 ) 
        	{
          		n += fread(readchunk3+n, 1, 4-(n&3), fh);
        	}
      	} 
      	else 
      	{
        	n = min2(sizeof(readchunk3), clen);
      	}
      	if( (n == 0) || ((n&3) != 0) ) 
      	{
        	if( (clen != 0) || (n != 0) ) 
        	{
          		fprintf(stderr, "ERROR: short read, checksum may be incorrect.\n");
        	}
        	break;
      	}
      	for( i=0; i<n; i+=4 ) 
      	{
	        c1 = BYTES2LONG(&readchunk3[i], swapped);
	        k1 = t6 + c1;
	        if( k1 < t6 ) t4++;
	        t6 = k1;
	        t3 ^= c1;
	        k2 = c1 & 0x1f;
	        k1 = ROL(c1, k2);
	        t5 += k1;
	        if( c1 < t2 ) 
	        {
	        	t2 ^= k1;
	        } 
	        else 
	        {
	        	t2 ^= t6 ^ c1;
	        }
	        t1 += c1 ^ t5;
      	}
      	if( rlen > 0 ) 
      	{
        	rlen -= n;
        	if( rlen <= 0 ) memset(readchunk3, 0, sizeof(readchunk3));
      	}
	clen -= n;
}
sum1 = t6 ^ t4 ^ t3;
sum2 = t5 ^ t2 ^ t1;

  LONG2BYTES(sum1, &readchunk3[0], 0);
  LONG2BYTES(sum2, &readchunk3[4], 0);
  fseek(fh, CHECKSUM_HEADERPOS, SEEK_SET);
  fread(readchunk3+8, 1, 8, fh);
//  printf("%02x %02x %02x %02x  ", readchunk3[ 8^swapped], readchunk3[ 9^swapped], readchunk3[10^swapped], readchunk3[11^swapped]);
//  printf("%02x %02x %02x %02x\n", readchunk3[12^swapped], readchunk3[13^swapped], readchunk3[14^swapped], readchunk3[15^swapped]);
//  printf("%02x %02x %02x %02x  ", readchunk3[0], readchunk3[1], readchunk3[2], readchunk3[3]);
//  printf("%02x %02x %02x %02x\n", readchunk3[4], readchunk3[5], readchunk3[6], readchunk3[7]);
  fseek(fh, CHECKSUM_HEADERPOS, SEEK_SET);
  LONG2BYTES(sum1, &readchunk3[16], swapped);
  LONG2BYTES(sum2, &readchunk3[20], swapped);
  fwrite(readchunk3+16, 1, 8, fh);
  fclose(fh);
fhexdump(name,CHECKSUM_HEADERPOS+hsize,8);
printf("\n");
	return(0);
}


#define MBIT 131072

//options

#define NFO	11//show rom info
#define CHK	5//fix checksum
#define PAD	6//fix size
#define V64	7//convert to doctor v64
#define Z64	8//convert to z64
#define PNF	3//palntsc fix
#define SWAP	9//swap bytes
#define XV64	21//send rom to doctor v64
#define XVJR	22//send rom to doctor v64
#define SRAM    10
#define NAME 	55//change name
#define BOT	12
#define USMS	13

//this single routine does all the nintendo64 stuff

int nintendo64(char *name,int option,char *option2)
{
//option2[0] = swapped: '1'=yes, ' '=no, '?'=unknown
//&option2[16] = string

#define N64START 0

char *maker[]={"0",\
"1",\
"2",\
"3",\
"4",\
"5",\
"6",\
"7",\
"8",\
"9",\
"10",\
"11",\
"12",\
"13",\
"14",\
"15",\
"16",\
"17",\
"18",\
"19",\
"20",\
"21",\
"22",\
"23",\
"24",\
"25",\
"26",\
"27",\
"28",\
"29",\
"30",\
"31",\
"32",\
"33",\
"34",\
"35",\
"36",\
"37",\
"38",\
"39",\
"40",\
"41",\
"42",\
"43",\
"44",\
"45",\
"46",\
"47",\
"48",\
"49",\
"50",\
"51",\
"52",\
"53",\
"54",\
"55",\
"56",\
"57",\
"58",\
"59",\
"60",\
"61",\
"62",\
"63",\
"64",\
"65",\
"66",\
"Nintendo",\
"68",\
"69",\
"70",\
"71",\
"72",\
"73",\
"74",\
"75",\
"76",\
"77",\
"Nintendo",\
"79",\
"80",\
"81",\
"82",\
"83",\
"84",\
"85",\
"86",\
"87",\
"88",\
"89",\
"90",\
"91",\
"92",\
"93",\
"94",\
"95",\
"96",\
"97",\
"98",\
"99",\
"100",\
"101",\
"102",\
"103",\
"104",\
"105",\
"106",\
"107",\
"108",\
"109",\
"110",\
"111",\
"112",\
"113",\
"114",\
"115",\
"116",\
"117",\
"118",\
"119",\
"120",\
"121",\
"122",\
"123",\
"124",\
"125",\
"126",\
"127",\
"128",\
"129",\
"130",\
"131",\
"132",\
"133",\
"134",\
"135",\
"136",\
"137",\
"138",\
"139",\
"140",\
"141",\
"142",\
"143",\
"144",\
"145",\
"146",\
"147",\
"148",\
"149",\
"150",\
"151",\
"152",\
"153",\
"154",\
"155",\
"156",\
"157",\
"158",\
"159",\
"160",\
"161",\
"162",\
"163",\
"164",\
"165",\
"166",\
"167",\
"168",\
"169",\
"170",\
"171",\
"172",\
"173",\
"174",\
"175",\
"176",\
"177",\
"178",\
"179",\
"180",\
"181",\
"182",\
"183",\
"184",\
"185",\
"186",\
"187",\
"188",\
"189",\
"190",\
"191",\
"192",\
"193",\
"194",\
"195",\
"196",\
"197",\
"198",\
"199",\
"200",\
"201",\
"202",\
"203",\
"204",\
"205",\
"206",\
"207",\
"208",\
"209",\
"210",\
"211",\
"212",\
"213",\
"214",\
"215",\
"216",\
"217",\
"218",\
"219",\
"220",\
"221",\
"222",\
"223",\
"224",\
"225",\
"226",\
"227",\
"228",\
"229",\
"230",\
"231",\
"232",\
"233",\
"234",\
"235",\
"236",\
"237",\
"238",\
"239",\
"240",\
"241",\
"242",\
"243",\
"244",\
"245",\
"246",\
"247",\
"248",\
"249",\
"250",\
"251",\
"252",\
"253",\
"254",\
"255",\
};

char *country[]={\
"0",\
"1",\
"2",\
"3",\
"4",\
"5",\
"6",\
"7",\
"8",\
"9",\
"10",\
"11",\
"12",\
"13",\
"14",\
"15",\
"16",\
"17",\
"18",\
"19",\
"20",\
"21",\
"22",\
"23",\
"24",\
"25",\
"26",\
"27",\
"28",\
"29",\
"30",\
"31",\
"32",\
"33",\
"34",\
"35",\
"36",\
"37",\
"38",\
"39",\
"40",\
"41",\
"42",\
"43",\
"44",\
"45",\
"46",\
"47",\
"48",\
"49",\
"50",\
"51",\
"52",\
"53",\
"54",\
"55",\
"56",\
"57",\
"58",\
"59",\
"60",\
"61",\
"62",\
"63",\
"64",\
"65",\
"66",\
"67",\
"Germany",\
"U.S.A.",\
"70",\
"71",\
"72",\
"73",\
"Japan",\
"75",\
"76",\
"77",\
"78",\
"79",\
"Europe",\
"81",\
"82",\
"83",\
"84",\
"Australia",\
"86",\
"87",\
"France, German, Dutch",\
"89",\
"90",\
"91",\
"92",\
"93",\
"94",\
"95",\
"96",\
"97",\
"98",\
"99",\
"100",\
"101",\
"102",\
"103",\
"104",\
"105",\
"106",\
"107",\
"108",\
"109",\
"110",\
"111",\
"112",\
"113",\
"114",\
"115",\
"116",\
"117",\
"118",\
"119",\
"120",\
"121",\
"122",\
"123",\
"124",\
"125",\
"126",\
"127",\
"128",\
"129",\
"130",\
"131",\
"132",\
"133",\
"134",\
"135",\
"136",\
"137",\
"138",\
"139",\
"140",\
"141",\
"142",\
"143",\
"144",\
"145",\
"146",\
"147",\
"148",\
"149",\
"150",\
"151",\
"152",\
"153",\
"154",\
"155",\
"156",\
"157",\
"158",\
"159",\
"160",\
"161",\
"162",\
"163",\
"164",\
"165",\
"166",\
"167",\
"168",\
"169",\
"170",\
"171",\
"172",\
"173",\
"174",\
"175",\
"176",\
"177",\
"178",\
"179",\
"180",\
"181",\
"182",\
"183",\
"184",\
"185",\
"186",\
"187",\
"188",\
"189",\
"190",\
"191",\
"192",\
"193",\
"194",\
"195",\
"196",\
"197",\
"198",\
"199",\
"200",\
"201",\
"202",\
"203",\
"204",\
"205",\
"206",\
"207",\
"208",\
"209",\
"210",\
"211",\
"212",\
"213",\
"214",\
"215",\
"216",\
"217",\
"218",\
"219",\
"220",\
"221",\
"222",\
"223",\
"224",\
"225",\
"226",\
"227",\
"228",\
"229",\
"230",\
"231",\
"232",\
"233",\
"234",\
"235",\
"236",\
"237",\
"238",\
"239",\
"240",\
"241",\
"242",\
"243",\
"244",\
"245",\
"246",\
"247",\
"248",\
"249",\
"250",\
"251",\
"252",\
"253",\
"254",\
"255",\
};
char *copier[]={"Z64","Doctor V64/V64Jr"};
register long x;
char sram[512];
char romname[4096];
char romheader[64];
char buf[32768];
char *usmsbuf;
//char buf2[4096];
unsigned long hsize=0;
//char header[512];
int swapped=0;
FILE *fh;
    
strcpy(romname,name);

//first indentify the rom as a nintendo64 rom and test for header

x=0;
x=x+fgetc2(romname,N64START+hsize)*256*256*256;
x=x+fgetc2(romname,N64START+hsize+1)*256*256;
x=x+fgetc2(romname,N64START+hsize+2)*256;
x=x+fgetc2(romname,N64START+hsize+3);
swapped=hsize=0;
if(x!=0x80371240)
{
	swapped=1;hsize=0;
	if(x!=0x37804012){if(!option)return(-1);}
}
if(fsize(romname,hsize)<32768){if(!option)return(-1);}

if(option2[0]!='?')swapped=((option2[0]=='0')?0:1);

switch(option)
{
case PAD:
//	strcpy(romname,fmkrom(romname,hsize));
	strcpy(buf,romname);
	fpad(buf,hsize,"wb");
break;

case V64:
//	strcpy(romname,fmkrom(romname,hsize));
	strcpy(buf,romname);
	buf[strcspn(buf,".")+1]=0;
	strcat(buf,isstrlwr(buf)?"v64":"V64");

	fcopy(romname,0,buf,"wb");
	if(!swapped)fswap(buf,0,fsize(buf,hsize));
break;

case XV64:
	x=transfer(romname,TRANSFER_V64,&option2[16],swapped);
	printf("\n");
        nintendo64(romname,NFO,option2);
        return(x);
break;

case XVJR:
	x=transfer(romname,TRANSFER_V64JR,&option2[16],swapped);
	printf("\n");
        nintendo64(romname,NFO,option2);
        return(x);
break;
	
case Z64:	
//	strcpy(romname,fmkrom(romname,hsize));
	strcpy(buf,romname);
	buf[strcspn(buf,".")+1]=0;
	strcat(buf,isstrlwr(buf)?"z64":"Z64");

	fcopy(romname,0,buf,"wb");
	if(swapped!=0)fswap(buf,0,fsize(buf,hsize));
break;

case SRAM:
//	This is the support of LaC's makesram routine for uploading
//	SRAM files to a Cart SRAM.  The .v64 file is his work, not mine

	if(fsize(&option2[16],0)!=512)return(-1);
	if(fsize(romname,0)!=1310720)return(-1);

	if(!(fh=fopen(&option2[16],"rb")))return(-1);
	fread(sram,sizeof(sram),1,fh);
	fclose(fh);
	        
	if(swapped!=0)fswap(sram,0,sizeof(sram));

	if(!(fh=fopen(romname,"r+b")))return(-1);
	fseek(fh,0x0286C0,SEEK_SET);//position of the sram in lac's rom file
	fwrite(sram,sizeof(sram),1,fh);
	fclose(fh);

	return(n64chksum(romname,hsize,swapped));
break;

case CHK:
	return(n64chksum(romname,hsize,swapped));//rewrite/fix checksum
break;

case SWAP:
	strcpy(buf,romname);
	buf[strcspn(buf,".")+1]=0;
	strcat(buf,isstrlwr(buf)?"rom":"ROM");

	fcopy(romname,0,buf,"wb");
//	if(!swapped)fswap(buf,0,fsize(buf,hsize));
	fswap(buf,0,fsize(buf,hsize));
break;

case BOT:
	if(option2[16]!=0&&!access(&option2[16],F_OK))
	{
		fstrncpy(buf,&option2[16],0,(sizeof(buf)<fsize(&option2[16],0))?sizeof(buf):fsize(&option2[16],0));
		if(swapped!=0)fswap(buf,0,0x01000-0x040);
		fhexdump(buf,0,0x01000-0x040);
		fstrncpy(romname,buf,N64START+hsize+0x040,0x01000-0x040);
		return(n64chksum(romname,hsize,swapped));
	}
	else
	{
		strcpy(buf,romname);
		buf[strcspn(buf,".")+1]=0;
		strcat(buf,isstrlwr(buf)?"bot":"BOT");
		fcopy3(romname,N64START+hsize+0x040,0x01000-0x040,buf,"wb");
		if(swapped!=0)fswap(buf,0,fsize(buf,0));
	}
break;

case USMS:
//must be smaller than 4mbit 524288 bytes will be inserted
//from 1b410 to 9b40f (7ffff)

	if(option2[16]!=0&&!access(&option2[16],F_OK))
	{
		if(fsize(&option2[16],0)>0x07fffff)return(-1);
		if(!(usmsbuf=(char *)malloc((fsize(&option2[16],0)+2)*sizeof(char))))return(-1);
		fstrncpy(usmsbuf,&option2[16],0,fsize(&option2[16],0));

		if(swapped!=0)fswap(usmsbuf,0,fsize(&option2[16],0));

		fcopy0(romname,"patched.v64");
		fstrncpy("patched.v64",usmsbuf,N64START+hsize+0x01b410,fsize(&option2[16],0));
		free(usmsbuf);
		return(n64chksum("patched.v64",hsize,swapped));
	}
	else return(-1);
//	{
//		strcpy(buf,romname);
//		buf[strcspn(buf,".")+1]=0;
//		strcat(buf,isstrlwr(buf)?"gg":"GG");
//		fcopy3(romname,N64START+hsize+0x040,0x01000-0x040,buf,"wb");
//		if(swapped!=0)fswap(buf,0,fsize(buf,0));
//	}
break;

case NAME:
	strcpy(buf,&option2[16]);
//        strcat(option2,"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0");
        	strcat(buf,"                     ");
	if(swapped!=0)fswap(buf,0,strlen(buf));
	fstrncpy(romname,buf,N64START+hsize+32,20);
	nintendo64(name,NFO,option2);
break;

case NFO:
	if(access(name,F_OK)!=0)return(0);//receive?
/*
#ifdef TR_ENDIAN_IS_LITTLE
        rominfo.validation           = *(tr_u16 *)(rom.header + (0x00 ^ 0x02));
        rominfo.compression          = *(tr_u8 *) (rom.header + (0x02 ^ 0x03));
        rominfo.unknown1             = *(tr_u8 *) (rom.header + (0x03 ^ 0x03));
        rominfo.clockrate            = *(tr_u32 *) (rom.header +  0x04);
        rominfo.programcounter       = *(tr_u32 *) (rom.header +  0x08);
        rominfo.release              = *(tr_u32 *) (rom.header +  0x0c);
        rominfo.crc1                 = *(tr_u32 *) (rom.header +  0x10);
        rominfo.crc2                 = *(tr_u32 *) (rom.header +  0x14);
        rominfo.unknown2_h           = *(tr_u32 *) (rom.header +  0x18);
        rominfo.unknown2_l           = *(tr_u32 *) (rom.header +  0x1c);
        for(i=0; i<20; i++)
                rominfo.name[i^0x03] = rom.header[i+0x20];
        rominfo.unknown3             = *(tr_u8 *) (rom.header + (0x34 ^ 0x03));
        rominfo.unknown4             = *(tr_u8 *) (rom.header + (0x35 ^ 0x03));
        rominfo.unknown5             = *(tr_u8 *) (rom.header + (0x36 ^ 0x03));
        rominfo.unknown6             = *(tr_u8 *) (rom.header + (0x37 ^ 0x03));
        rominfo.unknown7             = *(tr_u8 *) (rom.header + (0x38 ^ 0x03));
        rominfo.unknown8             = *(tr_u8 *) (rom.header + (0x39 ^ 0x03));
        rominfo.unknown9             = *(tr_u8 *) (rom.header + (0x3a ^ 0x03));
        rominfo.manufacturerid       = *(tr_u8 *) (rom.header + (0x3b ^ 0x03));
        rominfo.cartridgeid          = *(tr_u16 *)(rom.header + (0x3c ^ 0x02));
        rominfo.countrycode          = *(tr_u8 *) (rom.header + (0x3e ^ 0x03));
        rominfo.unknown10            = *(tr_u8 *) (rom.header + (0x3f ^ 0x03));
#endif
   
        rom.prgcode_base_orig = rominfo.programcounter;
        rom.prgcode_base = rom.prgcode_base_orig & 0x1fffffff;

        printf("ROM length      : 0x%08lx (%lu) bytes\n", rom.length, rom.length);
        printf("ROM valid. hword: %04hx\n",     rominfo.validation);
        printf("ROM comp. byte  : %01x (%s)\n",
                rominfo.compression,
                rominfo.compression == 0x12
                        ? "uncompressed"
                        : rominfo.compression == 0x13
                                ? "compressed"
                                : "?");
        printf("ROM ????????????: %01x\n",      rominfo.unknown1);
        printf("ROM clock rate  : %08"PR32"x\n",     rominfo.clockrate);
        printf("ROM pc          : %08"PR32"x\n",     rominfo.programcounter);
        printf("ROM release     : %08"PR32"x\n",     rominfo.release);
        printf("ROM crc 1       : %08"PR32"x\n",     rominfo.crc1);
        printf("ROM crc 2       : %08"PR32"x\n",     rominfo.crc2);
        printf("ROM name        : \"%s\"\n",      rominfo.name);
        printf("ROM manufac. id : %02x (%s)\n", rominfo.manufacturerid, manufacturerid[rominfo.manufacturerid]);
        printf("ROM cartridgeid : %04hx\n",     rominfo.cartridgeid);
        printf("ROM country code: %02x (%s)\n", rominfo.countrycode, countrycode[rominfo.countrycode]);
        puts("-------------------------------------------------------------------------------"); fflush(stdout);

*/


	fhexdump(romname,N64START+hsize,sizeof(romheader));
	printf("\n");
	printf("Nintendo 64\n");
	printf("%s\n",copier[(!swapped)?0:1]);

	for(x=0;x<20;x++)buf[x]=fgetc2(romname,N64START+x+32);
	buf[20]=0;
	if(swapped!=0)fswap(buf,0,strlen(buf));
	printf("%s\n",buf);

	printf("%s\n",maker[fgetc2(romname,N64START+hsize+((!swapped)?59:58))]);

	printf("%s\n",country[fgetc2(romname,N64START+hsize+((!swapped)?62:63))]);
	printf("%ld bytes (%.4f MBit)\n",fsize(romname,hsize),(float)fsize(romname,hsize)/MBIT);


printf("\n");


/*	if((x=fpad(romname,hsize,"rb"))!=-1)
	{
		if(!x)printf("Padded: No\n");
		else printf("Padded: Maybe, %ld bytes (%.4f MBit)\n",x,(float)x/MBIT);
	}
*/

/*	printf("%s\n",romtype[fgetc2(romname,N64START+hirom+hsize+38)]);
	printf((!hirom)?"":"HiRom\n");
*/
	x=(fsize(romname,hsize)-(fsize(romname,hsize)/MBIT)*MBIT);
	if(!x)printf("Intro/Trainer: No\n");
	else printf("Intro/Trainer: Yes, %ld bytes\n",x);
	printf("Swapped: %s\n",(!swapped)?"No":"Yes");
	printf("CheckSum: ");
	for(x=0;x<8;x++)
	{
		printf("%02x",fgetc2(romname,N64START+hsize+16+(x^swapped)));
		if(x==3)printf(" ");
	}
	printf("\n");
/*
	y=fgetc2(romname,N64START+hirom+hsize+40);
	for(x=1;x<y+2;x=x*2);
	x=x*8;
	printf((!x)?"":"%ldk Save RAM\n",x);
*/
//	printf("v1.%d\n",fgetc2(romname,N64START+hsize+((!swapped)?62:63));

	printf("\n");
break;

default:
break;
}
return(0);
}
